/*  Copyright 2009 Marc Toussaint
    email: mtoussai@cs.tu-berlin.de

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a COPYING file of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/> */

/*! \file util.h
    \brief many utilities: contains a bunch of standard includes, macros, operators,
    etc. that are used by most of my code */
#ifndef MT_util_h
#define MT_util_h

//----- if no system flag, I assume Linux
#if !defined MT_MSVC && !defined MT_Cygwin && !defined MT_Linux && !defined MT_MinGW
#  define MT_Linux
#endif

//----- standard includes:
#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <typeinfo>
#include <math.h>
#include <string.h>
#if defined MT_Linux || defined MT_Cygwin
#  include<sys/time.h>
#  include<sys/times.h>
#  include<sys/resource.h>
#endif
#if defined MT_MSVC
#  include<time.h>
#  include<sys/timeb.h> 
#  include<windows.h>
#  undef min
#  undef max
#  define MT_TIMEB
#  define __FUNCTION__ ""
#  ifdef MT_QT
#    undef  NOUNICODE
#    define NOUNICODE
#  endif
#  pragma warning(disable: 4305 4244 4250 4355 4786 4996)
#endif
#if defined MT_MinGW
#  include<unistd.h>
#  include<sys/time.h>
#  include<sys/timeb.h>
#  define MT_TIMEB
#endif


//----- declare my namespace for the first time:
//! Marc Toussaint namespace
namespace MT{
  inline void breakPoint(){
    int i;
    i=5;    //set a break point here, if you want to catch errors directly
  }
}


//----- error handling:
#ifdef MT_MSVC
#  define MT_HERE "@"<<(strrchr(__FILE__,'\\')?strrchr(__FILE__,'\\')+1:__FILE__)<<':'<<__LINE__<<':'<<__FUNCTION__<<": "
#else
#  define MT_HERE "@"<<(strrchr(__FILE__,'/')?strrchr(__FILE__,'/')+1:__FILE__)<<':'<<__LINE__<<':'<<__FUNCTION__<<": "
#endif
#ifndef MT_MSG
#  define MT_MSG(msg) { std::cerr <<MT_HERE <<msg <<std::endl; MT::breakPoint(); }
#endif
#ifndef HALT
#  define HALT(msg) { std::cerr <<MT_HERE <<msg <<" --- HALT" <<std::endl; MT::breakPoint(); throw STRING(msg); }
#  define NIY HALT("not implemented yet")
#endif


//----- check macros:
#ifndef MT_NOCHECK
#  define MT_DEBUG(x) x
#  define CHECK(cond,msg) if(!(cond)) HALT("CHECK failed: "<<msg);
#else
#  define MT_DEBUG(x)
#  define CHECK(cond,msg)
#endif


//----- basic defs:
#define MT_PI 3.14159265358979323846
#define MT_LN2 0.69314718055994528622676398299518041312694549560546875
#define MT_2PI 6.283195307179587
#define MT_LnSqrt2Pi -0.9189385332046727417803296 
#define MT_SQRT2 1.414213562373095049
#define MT_SQRTPI 1.772453850905516027
typedef unsigned char byte;            //!< byte
typedef unsigned short int uint16;     //!< 2 bytes
typedef unsigned int uint;             //!< unsigned integer
typedef unsigned long long int uint64; //!< 4 bytes


//----- macros to define the standard << and >> operatos for most my classes:
#define stdInPipe(type)\
  inline std::istream& operator>>(std::istream& is,type& x){ x.read(is);return is; }
#define stdOutPipe(type)\
  inline std::ostream& operator<<(std::ostream& os,const type& x){ x.write(os); return os; }
#define stdPipes(type)\
  inline std::istream& operator>>(std::istream& is,type& x){ x.read(is);return is; }\
  inline std::ostream& operator<<(std::ostream& os,const type& x){ x.write(os); return os; }\


//----- macros for piping doubles EXACTLY (without rounding errors) in hex coding:
#define OUTHEX(y) "0x" <<std::hex <<*((unsigned long*)&y) <<std::dec
#define INHEX(y)  std::hex >>*((unsigned long*)&y) >>std::dec


//===========================================================================
//
// standard little methods in my namespace
//

namespace MT{
  extern int argc;
  extern char** argv;
  extern std::ifstream cfgFile;
  extern bool cfgOpenFlag,cfgLock;
  extern bool IOraw;  //!< stream modifier for some classes (Mem in particular)
  extern bool noLog;  //!< no logfile: default=true, becomes false when MT::init is called
  extern uint lineCount;

  //----- files
  void open(std::ofstream& fs,const char *name,const char *errmsg="");
  void open(std::ifstream& fs,const char *name,const char *errmsg="");
  template<class T> void save(const T& x,const char *filename);
  template<class T> void load(T& x,const char *filename,bool change_directory=false);
  void decomposeFilename(char *&_path,char *&name,const char *filename);
  std::ofstream& log(const char *name="MT.log");
  
  //----- strings and streams
  bool contains(const char *s,char c);
  char skip(std::istream& is,const char *skipchars=" \n\r\t",bool skipCommentLines=true);
  void skipLine(std::istream& is);
  char peerNextChar(std::istream& is);
  void parse(std::istream& is,const char *str);
  bool skipUntil(std::istream& is,const char *tag);

  //----- functions
  byte bit(byte *str,uint i);
  void flip(byte& b,uint i);
  void flip(int& b,uint i);
  double modMetric(double x,double y,double mod);
  double sign(double x);
  double linsig(double x);
  void   constrain(double& x,double a,double b);
  double phi(double dx,double dy);
  double dphi(double x,double y,double dx,double dy);
  double DIV(double x,double y,bool force=false);
  double sigmoid11(double x);
  double sigmoid(double x);
  double dsigmoid(double x);
  double approxExp(double x);
  double Log(double x);
  uint   Log2(uint n);
  double sqr(double x);
  double sinc(double x);
  double cosc(double x);
  template<class T> T MIN(T a,T b){ return a<b?a:b; }
  template<class T> T MAX(T a,T b){ return a>b?a:b; }
  double erf(double x);
  double gaussInt(double x);
  double gaussIntExpectation(double x);

  //----- time access
  double realTime();
  double cpuTime();
  double sysTime();
  double totalTime();
  char *date();
  void wait(double sec);
  bool wait();

  //----- memory
  long mem();

  //----- timer functions
  void   timerStart();
  double timerRead(bool reset=false);
  double timerRead(bool reset,double startTime);
  double timerPause();
  void   timerResume();

  //----- command line handling
  void initCmdLine(int _argc, char *_argv[]);
  bool checkCmdLineTag(const char *tag);
  char *getCmdLineArgument(const char *tag);

  //----- config file handling
  void openConfigFile(const char *name=0);

  //----- parameter grabbing from command line, config file, or default value
  template<class T> T getParameter(const char *tag);
  template<class T> T getParameter(const char *tag,const T& Default);
  template<class T> void getParameter(T& x,const char *tag,const T& Default);
  template<class T> void getParameter(T& x,const char *tag);
  template<class T> bool checkParameter(const char *tag);
}

std::istream& operator>>(std::istream& is,const char *str);
std::istream& operator>>(std::istream& is,char *str);


//===========================================================================
//
// String class
//

#define STRING(x) (((MT::String&)(MT::String().stream()<<x)).p)

namespace MT{
  typedef std::iostream IOStream;
  
  /*!\brief String implements the functionalities of an ostream and an
  istream, but also can be send to an ostream or read from an
  istream. It is based on a simple streambuf derived from the
  MT::Mem class */
  class String:public IOStream{
    private:
      class StringBuf:public std::streambuf{
        public:
          String *string;
          virtual int overflow(int C = traits_type::eof());
          virtual int sync();
          void setIpos(char *p);
          char *getIpos();
      } buffer;
      void init();
      void append(char x);
      void resize(uint n,bool copy);
    
    public:
      //!@name data fields
      char *p;    //!< pointer to memory
      uint memN;  //!< \# elements (excluding zero)
      uint M;     //!< actual buffer size (in terms of # elements)
      static const char *readSkipSymbols; //!< default argument to read method (also called by operator>>)
      static const char *readStopSymbols; //!< default argument to read method (also called by operator>>)
      static int readEatStopSymbol;       //!< default argument to read method (also called by operator>>)
      void (*flushHandler)(String&);

      //!@name constructors
      String();
      String(const String& s);
      explicit String(const char *s);
      ~String();

      //!@name access
      uint N() const;
      operator char*();
      operator const char*() const;
      char &operator()(uint i) const;
      IOStream& stream();
      String& operator()();

      //!@name setting, appending
      String& operator=(const String& s);
      void operator=(const char *s);
      void set(const char *s,uint n);
      template<class T> String operator+(const T& v) const{ String news(*this); news <<v; return news; }
      template<class T> String prepend(const T& v) const{ String news; news <<v <<*this; return news; }
      
      //!@name resetting
      String& clr();
      String& resetI();

      //!@name equality
      bool operator==(const char *s);
      bool operator==(const String& s);

      //!@name I/O
      void write(std::ostream& os) const;
      void read(std::istream& is,const char* skipSymbols=NULL,const char *stopSymbols=NULL,int eatStopSymbol=-1);
  };
}
stdPipes(MT::String)


//===========================================================================
//
// a shared memory
//

namespace MT{

#define maxShmGuests 20
#define shmMaxBlocks 20
  struct SHMinfoBlock{
    const char pagename[20];
    uint size,used;
    uint guests;
    int guestPids[maxShmGuests];
    char blockNames[shmMaxBlocks][20];
    uint blockOffsets[shmMaxBlocks];
    pthread_rwlock_t blockLocks[shmMaxBlocks];
  };

  struct Lock{
  //documentation: http://cs.pub.ro/~apc/2003/resources/pthreads/uguide/document.htm
    pthread_rwlock_t lock;
    void init(bool process_shared=true){
      pthread_rwlockattr_t   att;
      int r;
      r=pthread_rwlockattr_init(&att); if(r) HALT("");
      if(process_shared){
        r=pthread_rwlockattr_setpshared(&att, PTHREAD_PROCESS_SHARED); if(r) HALT("");
      }
      r=pthread_rwlock_init(&lock, &att); if(r) HALT("");
    }
    void read(){
      int r=pthread_rwlock_rdlock(&lock);
      if(r) HALT("");
    }
  
    void write(){
      int r=pthread_rwlock_wrlock(&lock);
      if(r) HALT("");
    }
  
    void unlock(){
      int r=pthread_rwlock_unlock(&lock);
      if(r) HALT("");
    }
  };

  struct SHM{
    int systemId;
    uint guestId;
    byte *p;
    bool created,opened;
    SHMinfoBlock *info;
    SHM(){ created=opened=false; p=NULL; systemId=NULL; info=NULL; guestId=-1; }
    void open(const char *pagename,uint size,bool directDestroy=false);
    void close();
    void destroy();
    void readLock(uint i);
    void writeLock(uint i);
    void unlock(uint i);
    void report();
    template<class T> T* newBlock(const char* objName,uint *i=NULL);
  };

  template<class T> struct Shared{
    uint blockId;
    T *p;
    SHM *shm;
    Shared(SHM& _shm,const char* name){ shm=&_shm; p=shm->newBlock<T>(name,&blockId); }
    void readLock() { shm->readLock(blockId); }
    void writeLock(){ shm->writeLock(blockId); }
    void unlock()   { shm->unlock(blockId); }
    void set(const T& x){ writeLock(); *p=x; unlock(); }
    void get(T& x)      { readLock();  x=*p; unlock(); }
    T& operator()(){ return *p; }
  };
}

//===========================================================================
//
// Parameter class - I use it frequently to read parameters from file or cmd line
// 

#if 1
namespace MT{
  /*!\brief A parameter that initializes itself from the command line
    (use \c MT::init), parameter file, or a default value (priority in
    this order).  Initialization is done on the fly the _first_ time
    its value is queried (i.e., referenced by the cast operators).*/
  template<class type>
  class Parameter{
  public:
    const char *typeName;
    type value,Default;
    const char *tag;
    bool initialized,hasDefault;

  public:
    //!@name constructors

    //! Determines the tag to search for in parameter file/command line
    explicit Parameter(const char *_tag){
      typeName=typeid(type).name();
      //MT::parameters.append(this);
      initialized=false;
      tag=_tag;
      hasDefault=false;
    };

    /*!\brief specifies also a default value -- parameter does not have to but
      can be specified in the parameter file/command line */
    Parameter(const char *_tag,const type& _default){
      typeName=typeid(type).name();
      //MT::parameters.append(this);
      initialized=false;
      tag=_tag;
      hasDefault=true;
      Default=_default;
    };

    ~Parameter(){
      /*for(uint i=0;i<MT::parameters.N;i++){
	if(MT::parameters(i)==this){ MT::parameters.remove(i); return; }
      }*/
    }

    //!@name value access

    //! standard type conversion: returns a const of the parameter value
    operator type(){ if(!initialized) initialize(); return value; }

    //! ()-operator: returns an lvalue of the parameter value
    type& operator()(){ if(!initialized) initialize(); return value; }


    //!@name manipulation

    //! assigs a value to the parameter -- no further initialization needed
    type& operator=(const type v){ initialized=true; value=v; return value; }

    //! set the tag (replacing the one from the constructor)
    void setTag(char *_tag){ tag=_tag; }

    /*!\brief enforces that the parameter is reinitialized from the parameter
      file/command line, the next time it is referenced -- even if it
      has been initialized before */
    void reInitialize(){ initialized=false; }


    //!@name explicit grabbing


  private:
    void initialize();
  };

}
#endif


//===========================================================================
//
// random number generator
//

namespace MT{
  /*!\brief A random number generator. An global instantiation \c
    MT::rnd of a \c Rnd object is created. Use this one object to get
    random numbers.*/
  class Rnd{
  private:
    bool ready;
    int rpoint;     /* Feldindex    */
    long rfield[256];   /* Schieberegisterfeld  */


  public:
    //! ...
    Rnd(){ ready=false; };


  public://!@name initialization
    //! initialize with a specific seed
    unsigned long seed(unsigned long n);

    //! use Parameter<uint>("seed") as seed
    unsigned long seed();

    //! uses the internal clock to generate a seed
    unsigned long clockSeed();

  public://!@name access
    //! a initeger random number uniformly distributed in [0,?]
    uint num(){ if(!ready) seed(); return (uint)rnd250() >> 5; }
    //! same as \c num()
    uint operator ()(){ return num(); }
    //! a initeger random number uniformly distributed in [0,\c i-1]
    uint num(uint limit){
      CHECK(limit,"zero limit in rnd.num()"); return num() % limit; }
    uint num(int lo,int hi){ return lo+num(hi-lo+1); }
    //! same as \c num(i)
    uint operator ()(uint i){ return num(i); }
    uint operator ()(int lo,int hi){ return num(lo,hi); }
    //! a random variable uniformly distributed in [0,1]
    double uni(){ return ((double)num(1<<22))/(1<<22); }
    //! a random variable uniformly distributed in [\c low, \c high]
    double uni(double low,double high){ return low+uni()*(high-low); }
    //! a gaussian random variable with mean zero
    double gauss();
    /*!\brief a positive integer drawn from a poisson distribution with given
      \c mean; is case \c mean>100, a (positive) gauss number
      \c floor(mean+gauss(sqrt(mean))+.5) is returned */
    uint poisson(double mean);
    // @}


  private:
    long  rnd250(){
      rpoint = (rpoint+1) & 255;          // Index erhoehen
      return rfield[rpoint] =  rfield[(rpoint-250) & 255]
	                     ^ rfield[(rpoint-103) & 255];
    }

    void seed250(long seed);
  };

  //! The global Rnd object
  extern Rnd rnd;
}


//===========================================================================
//
// generic any container
//

struct Any{
  char *tag;
  const char *type;
  void *p;
  uint n;
  char delim;
  virtual ~Any(){};
  //virtual void read(std::ostream &os) = 0;
  virtual void write(std::ostream &os) const = 0;
  virtual void free() = 0;
  virtual Any *new_clone() = 0;
  //virtual void operator=(const Any&) = 0;
  //virtual void set(const void *_p,uint _n) = 0;
};
stdOutPipe(Any)

template<class T> Any* anyNew(const char* tag,const T &x);
template<class T> Any* anyNew(const char* tag,const T *x,uint n,char delim);


//===========================================================================
//
// gnuplot calls
//

void gnuplot(const char *command,const char* EPSfile=NULL);
void gnuplotClose();


//===========================================================================
//
// implementations
//

#if defined MT_IMPLEMENTATION | defined MT_IMPLEMENT_TEMPLATES
#  include "util_t.cpp"
#endif

#ifdef MT_IMPLEMENTATION
#  include"util.cpp"
#endif



//===========================================================================
//
// USING's
//

using std::cout;
using std::cerr;
using std::endl;
using std::flush;
using std::ostream;
using std::istream;
using std::ofstream;
using std::ifstream;
using MT::rnd;
using MT::String;


#endif
